const types = {
  COMMENT: "comment",
  CONTINUED: "continued",
  DATE: "date",
  GENERATION: "basegen",
  LINK: "link",
  PERSON: "person",
  SECTION: "section",
  TITLE: "title",
};

const parseLine = (line) => {
  const trimmedLine = (line ?? "").trim();
  if (!trimmedLine) return null;

  // Fields
  const { field, value } =
    trimmedLine.match(/(?<field>.+?): (?<value>.+)/)?.groups ?? {};
  if (field && value) {
    return { type: field, value };
  }

  // Comments
  if (trimmedLine.startsWith("#")) {
    return { type: types.COMMENT };
  }

  // People
  const { indicator, lineData } =
    trimmedLine.match(/(?<indicator>[+->]) (?<lineData>.+)/)?.groups ?? {};
  if (indicator && lineData) {
    const numSpaces = line.length - trimmedLine.length;
    const generation = Math.floor(numSpaces / 2);

    const isSpouse = indicator === "+";
    const isLink = indicator === ">";

    const properties = lineData
      .split("|")
      .map((p) => p.trim())
      .filter((p) => p);
    const [name, ...rest] = properties;

    if (name === "@continued") {
      return {
        type: types.CONTINUED,
        generation,
      };
    }

    if (isLink) {
      return {
        type: types.LINK,
        generation,
        name,
      };
    }

    return {
      type: types.PERSON,
      generation,
      spouse: isSpouse,
      name,
      properties: rest,
    };
  }
};

const parseInput = (input) => {
  const lines = input.split("\n");
  const data = {
    baseGeneration: 1,
    lines: [],
  };

  const lineCommands = lines
    .map((line) => parseLine(line))
    .filter((line) => line);

  lineCommands.forEach((command) => {
    if (command.type === types.TITLE) {
      data.title = command.value;
    }

    if (command.type === types.SECTION) {
      data.section = command.value;
    }

    if (command.type === types.GENERATION) {
      data.baseGeneration = parseInt(command.value);
    }

    if (command.type === types.DATE) {
      data.date = command.value;
    }

    if (command.type === types.COMMENT) {
      return;
    }

    if (command.type === types.PERSON) {
      data.lines.push({
        ...command,
        generation: command.generation + 1,
        displayGeneration: data.baseGeneration + command.generation,
      });
    }

    if (command.type === types.CONTINUED) {
      data.lines.push({
        ...command,
        generation: command.generation + 1,
        displayGeneration: data.baseGeneration + command.generation,
      });
    }

    if (command.type === types.LINK) {
      data.lines.push({
        ...command,
        generation: command.generation + 1,
      });
    }
  });

  return data;
};

const getClasses = (classList) => classList.filter((cls) => cls).join(" ");

const renderPersonProperties = (line) => (line.properties ?? []).join(" — ");

const renderIndicator = (line) => {
  const classes = getClasses(["indicator", line.spouse ? "spouse" : null]);

  let indicatorContent = "";
  if (line.type === types.LINK) {
    indicatorContent = "&rarr;";
  } else if (!line.spouse) {
    indicatorContent = `[${line.displayGeneration}]`;
  }

  return `<span class="${classes}">${indicatorContent}</span>`;
};

const renderName = (line) =>
  `<span class="name">${line.spouse ? "+ " : ""}${line.name}</span>`;

const renderLine = (line) => {
  const properties = renderPersonProperties(line);
  const indicatorHTML = renderIndicator(line);
  const nameHTML = renderName(line);

  const classes = getClasses([
    "line",
    `gen-${line.generation}`,
    line.spouse ? "spouse" : null,
    line.type === types.CONTINUED ? "continued" : null,
    line.type === types.LINK ? "link" : null,
  ]);

  if (line.type === types.CONTINUED) {
    return `<div class="${classes}">${indicatorHTML} <span class="label">Continued on next page...</span></div>`;
  }

  if (line.type === types.LINK) {
    return `<div class="${classes}">${indicatorHTML} <span class="label">${line.name}</span></div>`;
  }

  return `<div class="${classes}">${indicatorHTML} <span class="label">${nameHTML} ${
    properties ? `<span class="properties">&mdash; ${properties}</span>` : ""
  }</span></div>`;
};

const renderChart = (data) => {
  let output;

  const sectionHTML = data.section
    ? `<div class="section">${data.section}</div>`
    : null;
  const titleHTML = data.title ? `<h1>${data.title}</h1>` : null;
  const dateHTML = data.date ? `<h2 class="date">${data.date}</h2>` : null;

  const headerHTML = `<header>${[sectionHTML, titleHTML, dateHTML]
    .filter((h) => h)
    .join(" ")}</header>`;

  const peopleLinesHTML =
    (data.lines ?? []).map((line) => renderLine(line)).join("\n") ?? null;

  return {
    title: data.title,
    output: [
      sectionHTML || titleHTML || dateHTML ? headerHTML : null,
      peopleLinesHTML,
    ]
      .filter((h) => h)
      .join("\n"),
  };
};

onmessage = (e) => {
  const chartData = parseInput(e.data);
  const { output, title } = renderChart(chartData);

  postMessage({ output, title });
};
